#! /usr/bin/python3
# -*- coding=utf-8 -*-

# Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS

# This software is published by the license of CPU-OS License, you can use and 
# distibuted this software under this License. See CPU-OS Liense for more detail.

# You should have receive a copy of CPU-OS License. If not, please contact us 
# by email <support_os@cpu-os.ac.cn>  

import apt
import apt_pkg
import aptdaemon
from aptdaemon import client
import os

import software_center
from software_center import AptDatabase
from PkgCore import utils

import logging
import urllib.request, urllib.parse, urllib.error

global source_server
#source_server = 'http://repos.01os.org/'
source_server = 'https://update.os.nfschina.com/static/'

class FetchProgress(apt.progress.base.AcquireProgress):
    def __init__(self, pkg_name="", pkg_version="", send_download_signal_cb=None):
        self._cancel = False
        self.percent = 0.0
        self.required_download = -1
        self.send_download_signal_cb = send_download_signal_cb
        self.pkg_name = pkg_name
        self.pkg_version = pkg_version

    def send_download_signal(self, status, process, err_reason=""):
        if self.send_download_signal_cb is None:
            logging.debug("no send_download_signal in FetchProgress")
            return
        self.send_download_signal_cb(self.pkg_name, self.pkg_version, status, int(process), err_reason)

    def done(self, item):
        pass

    def fail(self, item):
        self._cancel = True
        logging.error(str(item))

    def pulse(self, owner):
        if self._cancel :
            return False # 取消下载
        if self.required_download == -1:
            self.percent = self.fetched_bytes * 100 / self.total_bytes
        elif self.required_download <= self.total_bytes:
            completed = self.total_bytes - self.required_download
            self.percent = (self.fetched_bytes + completed) * 100 / self.total_bytes
            logging.debug('percent: %d, required_download: %d, total_bytes: %d, fetched_bytes: %d'\
                % (self.percent, self.required_download, self.total_bytes, self.fetched_bytes))
        elif self.total_bytes == 0:
            self.percent = 100

        if int(self.percent) == 100:
            return True
        self.send_download_signal(status="inprogress", process=self.percent, err_reason="")
        return True # 继续下载

    def cancel(self):
        self._cancel = True

class AptAction():
    def __init__(self, func, cache_func):
        self.apt_client = client.AptClient()
        self.all_app = []
        self.all_app_info = {} # pkg, appname, icon_path, ...
        self.update_app_info()#获取软件中心软件信息存储到self.all_app_info中.
        self.callback = func
        self.cache_callback = cache_func
        self.cache = None

    def init_data(self):
        self.transaction = None
        self.action = ""
        self.error_details = ""
        self.old_env = {}
        if self.cache is None:
            self.cache=apt.Cache()
        else:
            self.cache.open()

    def get_all_app_info(self):
        return self.all_app_info

    def get_all_package_names(self):
        return list(self.all_app_info.keys())

    def get_app_info(self, pkgs):
        ret = {}
        for pkg in pkgs:
            if ( not pkg) or ( not pkg.name):
                continue
            app = None
            try:
                app = self.all_app_info[pkg.name]
                if not pkg.candidate:
                    app["size"] = 0
                else:
                    app["size"] = utils.humanize_size(getattr(pkg.candidate, "size", 0))
            except Exception as e:
                logging.info(str(e))
            if app is not None:
                ret[app['Package']] = app
        return ret

    #获取应用信息，主要通过软件中心数据库获取.
    def update_app_info(self):
        logging.info("update app info")
        self.all_app_info = self._list_all_app_info()

    def install(self, env, pkgs):
        if not self.check_network():
            msg = { 'status': 'failed', 'percent': 0, 'error': "download failed" }
            if self.callback is not None:
                self.callback(msg)
            return
        logging.debug("install Pkgs")
        self.init_data()
        self.action = "install"
        self._backup_env(env)
        try:
            #通过aptdaemon安装pkgs
            self.transaction = self.apt_client.install_packages(pkgs)
            #监视progress-changed、finished、error信号.
            self.transaction.connect("progress-changed", self.change_progress)
            self.transaction.connect("finished", self.finished)
            self.transaction.connect("error", self.on_error)
            self.transaction.run()
        except Exception as e:
            msg = { 'status': 'failed', 'percent': 0, 'error': "install error" }
            logging.error(str(e))
        finally:
            self._restore_env()

    def upgrade(self, env, pkgs):
        if not self.check_network():
            msg = { 'status': 'failed', 'percent': 0, 'error': "download failed" }
            if self.callback is not None:
                self.callback(msg)
            return
        logging.debug("upgrade Pkgs")
        self.init_data()
        self.action = "upgrade"
        self._backup_env(env)
        try:
            self.transaction = self.apt_client.upgrade_packages(pkgs)
            self.transaction.connect("progress-changed", self.change_progress)
            self.transaction.connect("finished", self.finished)
            self.transaction.connect("error", self.on_error_upgrade)
            self.transaction.run()
        except Exception as e:
            msg = { 'status': 'failed', 'percent': 0, 'error': "upgrade error" }
            logging.error(msg)
        finally:
            self._restore_env()

    def commit(self, env, pkgs_install, pkgs_upgrade, worker=None):
        if not self.check_network():
            msg = { 'status': 'failed', 'percent': 0, 'error': "download failed" }
            if self.callback is not None:
                self.callback(msg)
            return
        logging.debug("commit Pkgs && update")
        #self.update_cache(env)
        self.init_data()
        self.action = "commit"
        self._backup_env(env)
        reinstall = remove = purge = downgrade = []
        try:
            self.transaction = self.apt_client.commit_packages(
                pkgs_install, reinstall, remove, purge, pkgs_upgrade,
                downgrade)
            self.transaction.connect("progress-changed", self.change_progress, worker)
            self.transaction.connect("finished", self.finished, worker)
            self.transaction.connect("error", self.on_error, worker)
            self.transaction.connect("config-file-conflict", self.on_config_file_conflict, worker)
            self.transaction.run()
        except Exception as e:
            #msg = { 'status': 'failed', 'percent': 0, 'error': "commit error" }
            msg = { 'status': 'failed', 'percent': 0, 'error': str(e) }
            logging.error(msg)
            if self.callback is not None:
                self.callback(msg)
        finally:
            self._restore_env()

    def download_pkg(self, env, pkg_name, pkg_version, send_download_signal_cb):
        if not self.check_network():
            send_download_signal_cb( pkg_name, pkg_version, "failed", 0, "The network is unreachable.")
            send_download_signal_cb( pkg_name, pkg_version, "end", 0, "")
            return
        logging.debug("download pkgs")
        self.init_data()
        self.action = "download"
        self._backup_env(env)
        try:
            fprogress = FetchProgress(pkg_name, pkg_version, send_download_signal_cb)
            pm = apt_pkg.PackageManager(self.cache._depcache)
            fetcher = apt_pkg.Acquire(fprogress)

            if not pkg_name in self.cache:
                send_download_signal_cb( pkg_name, pkg_version, "failed", 0, "pkg is not in cache.")
                send_download_signal_cb( pkg_name, pkg_version, "end", 0, "")
                return
            pkg_download_mark_flag = False
            pkg = self.cache[pkg_name]
            if pkg_version == "":
                pkg.mark_install()
                pkg_download_mark_flag = True
            else:
                ver_list = pkg._pkg.version_list
                for ver in ver_list:
                    if ver.ver_str == pkg_version:
                        self.cache._depcache.set_candidate_ver(pkg._pkg, ver)
                        pkg.mark_install()
                        pkg_download_mark_flag = True

            # if pkg version can be download
            if pkg_download_mark_flag == False:
                send_download_signal_cb(pkg_name, pkg_version, "failed", 0, "this version of pkg can't be download")
            # if already downloaded
            elif self.cache.required_download == 0:
                send_download_signal_cb(pkg_name, pkg_version, "success", 100, "")
            else:
                self.cache._fetch_archives(fetcher, pm)
                send_download_signal_cb(pkg_name, pkg_version, "success", 100, "")

        except Exception as e:
            logging.error("download_pkg Exception:" + str(e))
            send_download_signal_cb( pkg_name, pkg_version, "failed", 0, str(e))
        finally:
            self._restore_env()
            send_download_signal_cb( pkg_name, pkg_version, "end", 0, "")

    #从软件中心数据库返回所有软件的软件包名，名称，版本，简述和详细描述，图标.
    def _list_all_app_info(self):
        result = {}
        db = AptDatabase()
        for app in db.list_all_softwares():
            package_name = app['Package']
            # 获取图标
            icon = software_center.get_icon_from_software(package_name)
            app['Icon'] = icon
            result[package_name] = app
        return result

    def _list_all_package_names(self):
        db = AptDatabase()
        packages = db.list_software_names()
        return packages

    def _backup_env(self, env):
        for key in env:
            oldval=None
            try:
                oldval = os.environ[key]
            except Exception as e:
                pass
            self.old_env[key]=oldval
            os.environ[key] = env[key]

    def _restore_env(self):
        for key in self.old_env:
            if self.old_env[key] is not None :
                os.environ[key] = self.old_env[key]
            else:
                del os.environ[key]

    def update_cache(self, env):
        if not self.check_network():
            msg = { 'status': 'failed', 'percent': 0, 'error': "download failed" }
            if self.cache_callback is not None:
                self.cache_callback(msg)
            return
        try:
            apt_pkg.pkgsystem_unlock()
        except SystemError:
            pass

        logging.debug("update_cache")
        self.init_data()
        self.action = "update_cache"
        self._backup_env(env)
        try:
            self.transaction = self.apt_client.update_cache()
            self.transaction.connect("finished", self.on_cache_finished)
            self.transaction.connect("error", self.on_cache_error)
            self.transaction.run()
        except Exception as e:
            msg = { 'status': 'failed', 'percent': 0, 'error': str(e) }
            if self.callback is not None:
                self.callback(msg)
            logging.error(msg)
        finally:
            self._restore_env()

    def change_progress(self, tr, pr, worker=None):
        msg = { 'status': 'inprogress', 'percent': pr, 'op': '', 'error':'' }
        if self.callback is not None:
            self.callback(msg)
        logging.debug(msg)
        if (worker is not None) and ("callback" in worker) and (worker["callback"] is not None):
            worker["callback"](worker["pkg_name"], worker["pkg_version"], "inprogress", pr, "")

    def on_cache_finished(self, tr, exit_status):
        logging.debug("on_cache_finished")
        logging.debug(exit_status)
        msg = ''
        if exit_status == 'exit-cancelled':
            if self.action == 'cancel':
                msg = {'status': 'success', 'percent': 0, 'error':''}
            if self.action == 'pause':
                msg = {'status': 'success', 'percent': 0, 'error':''}
        if exit_status == 'exit-success':
            msg = {'status': 'success', 'percent': 100, 'error':''}
        if exit_status == 'exit-failed':
            msg = {'status': 'failed', 'percent': 0, 'error':self.error_details }
        logging.debug(msg)
        if self.cache_callback is not None:
            self.cache_callback(msg)

    def on_cache_error(self, tr, error_code, error_details):
        logging.error("on_cache_error:" + str(error_code) + " " + str(error_details))
        msg = {'status': 'failed', 'percent': 0, 'error':self.error_details }
        logging.debug(msg)
        if self.cache_callback is not None:
            self.cache_callback(msg)

    def finished(self, tr, exit_status, worker=None):
        logging.debug("finished")
        logging.debug(exit_status)
        msg = ''
        if exit_status == 'exit-cancelled':
            if self.action == 'cancel':
                msg = {'status': 'success', 'percent': 0, 'error':''}
            if self.action == 'pause':
                msg = {'status': 'success', 'percent': 0, 'error':''}
        if exit_status == 'exit-success':
            msg = {'status': 'success', 'percent': 100, 'error':''}
        if exit_status == 'exit-failed':
            msg = {'status': 'failed', 'percent': 0, 'error':self.error_details }
        logging.debug(msg)
        if self.callback is not None:
            self.callback(msg)
        # worker callback if it exist
        if (worker is not None) and ("callback" in worker) and (worker["callback"] is not None):
            if exit_status == "exit-failed":
                worker["callback"](worker["pkg_name"], worker["pkg_version"], "failed", 0, self.error_details)
            elif exit_status == "exit-success":
                worker["callback"](worker["pkg_name"], worker["pkg_version"], "success", 100, "")
            else:
                worker["callback"](worker["pkg_name"], worker["pkg_version"], "success", 0, "")

    def on_error_upgrade(self, tr, error_code, error_details):
        logging.error(error_code)
        if error_code == 'error-package-not-installed':
            self.error_details = 'upgradeWrong'
        else:
            self.error_details = "failed"
        logging.error(self.error_details)

    def on_error(self, tr, error_code, error_details, worker=None):
        logging.error(error_code)
        if error_code == 'error-package-download-failed':
            self.error_details = "download failed"
        elif error_code == 'error-repo-download-failed':
            self.error_details = "repo update failed"
        elif error_code == 'error-dep-resolution-failed':
            self.error_details = "dep failed"
        elif error_code == 'error-key-not-installed':
            self.error_details = "key no install"
        elif error_code == 'error-key-not-removed':
            self.error_details = "key no remove"
        elif error_code == 'error-no-lock':
            self.error_details = "no lock"
        elif error_code == 'error-no-cache':
            self.error_details = "no cache"
        elif error_code == 'error-no-package':
             self.error_details = "no package"
        elif error_code == 'error-package-uptodate':
            self.error_details = "not uptodate"
        elif error_code == 'error-package-not-installed':
            self.error_details = "package no install"
        elif error_code == 'error-package-already-installed':
            self.error_details = "already install"
        elif error_code == 'error-not-remove-essential':
            self.error_details = "essential"
        elif error_code == 'error-daemon-died':
            self.error_details = "daemon destroy"
        elif error_code == 'error-package-manager-failed':
            self.error_details = "manager failed"
        elif error_code == 'error-cache-broken':
             self.error_details = "cache broken"
        elif error_code == 'error-package-unauthenticated':
            self.error_details = "unauthenticated"
        elif error_code == 'error-incomplete-install':
            self.error_details = "uninstall"
        elif error_code == 'error-unreadable-package-file':
            self.error_details = "unreadable package file"
        elif error_code == 'error-invalid-package-file':
            self.error_details = "invalid package"
        elif error_code == 'error-system-already-uptodate':
             self.error_details = "not update"
        elif error_code == 'error-not-supported':
            self.error_details = "no support"
        elif error_code == 'error-license-key-install-failed':
            self.error_details = "license key error"
        elif error_code == 'error-license-key-download-failed':
            self.error_details = "license key download failed"
        elif error_code == 'error-unknown':
            self.error_details = "unknown"
        else:
            self.error_details = "failed"
        logging.error(self.error_details)

    def on_config_file_conflict(self, transaction, old, new, worker=None):
        logging.info("config_file_conflict:" + old)
        transaction.resolve_config_file_conflict(old, "keep")
        #transaction.resolve_config_file_conflict(old, "replace")
        logging.info("==========after config file conflict=========")

    #网络检查，可能需要修改？？？
    def check_network(self):
        # 无法检测404异常，仅用来检测是否有网络链接
        res = None
        try:
            global source_server
            res = urllib.request.urlopen(source_server)
        except Exception as e:
            logging.error(str(e))
            return False
        content = str(res.read())
        if content == 'fail':
            return False
        return True

if __name__ == '__main__':
    from gi.repository import GLib
    apt_action = AptAction(None, None)
    #print( apt_action.get_app_info(['altos', 'code','kazam']) )

    env = { 'HOME': os.environ['HOME'], 'USER': os.environ['USER'], 'DISPLAY':os.environ['DISPLAY'] }
    #apt_action.install( env, ['vim'])

    print((apt_action.check_network()))

    GLib.MainLoop().run()

